dofile("math.lua")

declare("TurretViewer", {})

local function makeangle(x, y, z)
	return math3d.MakeQuatFromGVectors(gvector(0, 0, 1), gvector(x, y, z))
end

local function rotateq(x, y, z, q)
	local q1 = makeangle(x, y, z)
	q1:rotate(q)
	return q1
end

local abs = math.abs
local cut = 1e-03

function TurretViewer.Create(_turrettab)
	local turretviewer = {}

	turretviewer.meshfiles = { -- there's no way to reliably get a mesh filepath without having the ship in your inventory...
		["ship_frigate_001"] = "gaf/ship_frigate_001.dfs", -- HAC
		["ship_trident_01"] = "gaf/ship_trident_01.dfs",
		["ship_tpg_capship"] = "gaf/ship_tpg_capship.lod", -- teradon
		["ship_tpg_capship2"] = "gaf/ship_tpg_capship2.dfs", -- constellation
		["ship_generic_behemoth"] = "gaf/ship_generic_behemoth.dfs",
		["ship_generic_transport"] = "gaf/ship_generic_transport.dfs", -- atlas
		["ship_leviathon_curved"] = "gaf/ship_leviathan.dfs",
	}

	function turretviewer.GetMeshName(shipname) -- shipname being the game's internal names for objects
		local invid = InvManager.InventoryItemToType(shipname)
		if not invid then return end
		local prototype = InvManager.GetInventoryItemPrototype(invid)
		if prototype then
			return prototype:GetMeshName()
		end
	end

	function turretviewer:GetMeshInfo(shipname)
		local meshname = self.GetMeshName(shipname)
		local meshpath
		if meshname then
			meshpath = self.meshfiles[meshname] or "gaf/"..meshname..".dfs" -- got to guess if we don't have it cached...
		end
		return meshname, meshpath
	end

	turretviewer.meshinfo = {}

	turretviewer.showturrets = true
	turretviewer.reset = true
	turretviewer.visible = false

	turretviewer.views = {
		{name="Top", angle=makeangle(0, 1, 0), cutoff=function(v) return v:y() >= 0 end},
		{name="Bottom", angle=rotateq(0, 1, 0, quaternion(0, 0, 1, 0)), cutoff=function(v) return v:y() < 0 end},
		{name="Right", angle=makeangle(1, 0, 0), cutoff=function(v) return v:x() >= -cut end},
		{name="Left", angle=makeangle(-1, 0, 0), cutoff=function(v) return v:x() <= cut end},
		{name="Front", angle=makeangle(0, 0, -1), cutoff=function(v) return v:z() >= cut end},
		{name="Back", angle=makeangle(0, 0, 1), cutoff=function(v) return v:z() < -cut end},
	}

	function turretviewer:CreateModelViewer()
		local this = self
		
		local iconcontainer = iup.cbox{active="NO", expand="YES"}
		
		local name = iup.label{title=""}
		
		local modelview = iup.modelview{value="", expand="YES", mode=1, active="NO"}
		modelview.timer = Timer()
		
		local box = iup.zbox{modelview, iconcontainer, all="YES", alignment="NW"}
		
		function modelview:SlerpOrientation(targetangle)
			iconcontainer.visible = this.showturrets and "YES" or "NO"
			
			if not self.currentangle then
				self:SetOrientation(targetangle)
				self.currentangle = targetangle
				return
			end
			
			
			if self.timer:IsActive() then
				if self.deltaangle then
					iconcontainer.visible = "NO"
					self.currentangle = self.deltaangle
				end
				self.timer:Kill()
			end
			
			if self.currentangle == targetangle then return end
			
			self.targetangle = targetangle
			
			local totaltime = 500 -- ms
			local sleeptime = math.floor(1000/60 + 0.5)
					
			local starttime = gkmisc.GetGameTime()
			
			local function timerfunc()
				local curtime = gkmisc.GetGameTime()
				local delta = (curtime-starttime)/totaltime			
				
				if delta < 1 then
					iconcontainer.visible = "NO"
					local angle = math3d.Slerp(self.currentangle, self.targetangle, math.smoothstep(0, 1, delta))
					self.deltaangle = angle
					self:SetOrientation(angle)
					self.timer:SetTimeout(sleeptime)
				else
					box:UpdateIconPositions()
					iconcontainer.visible = this.showturrets and "YES" or "NO"
					self:SetOrientation(self.targetangle)
					self.currentangle = self.targetangle
					self.targetangle = nil
				end
			end
			
			self:SetOrientation(self.currentangle)
			
			self.timer:SetTimeout(sleeptime, timerfunc)
		end
		
		local function getclosestsize(size)
			local sizes = {36, 34, 32, 28, 24, 20, 16}
			
			for i,v in ipairs(sizes) do
				if size >= v then
					return v
				end
			end
			
			return sizes[#sizes]
		end
		
		
		function box:SetupIcons()
			self.iconlist = self.iconlist or {}
			
			for k,v in pairs(self.iconlist) do
				HideTooltip()
				v.iconcontainer:detach()
				v.iconcontainer:destroy()
				self.iconlist[k] = nil
			end
			
			local n = GetNumStationTurrets()
			
			local iconlist = self.iconlist
			
			for i=1, n do
				local turretinfo = GetStationTurretInfo(i)
				-- this.turretlist[i] = turretinfo
				if turretinfo then
					local scale = gkinterface.GetYResolution()/600
					
					local turretdata = {}
					turretdata.info = turretinfo
					if n >= 16 then
						turretdata.size = 16
					elseif n >= 8 then
						turretdata.size = 20
					elseif n >= 4 then
						turretdata.size = 24
					else
						turretdata.size = 28
					end
					
					turretdata.size = getclosestsize(turretdata.size * scale)
					
					local iconsize = turretdata.size.."x"..turretdata.size
					
					local turretimage = "images/icon_ship_turret.png"
					
					turretdata.icon = iup.label{title="", size=iconsize, image=turretimage, bgcolor="255 255 255"}
					
					turretdata.selimage = iup.label{
						title="",
						image="images/icon_powerupselect.tga",
						size=iconsize,
						active="NO",
						visible="NO",
						bgcolor="255 255 255 255 +",
					}
					
					turretdata.button = iup.canvas{
						border="NO",
						expand="NO",
						size=iconsize,
						button_cb = function(self, button, pressed, x, y, status)
							if turretdata.info.isdead then return end
							if pressed == 1 then
								if button == iup.LMBUTTON then -- left click
									this.selectedturretdata = turretdata
									this.selectbutton.active = "YES"
									for i, modelview in pairs(this.modelviews) do
										for k,v in pairs(modelview.iconlist) do
											if v.info.itemid == turretdata.info.itemid then
												v.selimage.visible = "YES"
											else
												v.selimage.visible = "NO"
											end
										end
									end
								elseif button == iup.MBUTTONDBLCLICK or button == 8388616 then -- double click
									PurchaseMerchandiseItem(turretdata.info.itemid)
								end
							end
						end,
						enterwindow_cb = function(self)
							local info = turretdata.info
							local tooltiptext = ""
							local name = this.turretnames[info.itemid]
							if name then
								tooltiptext = name
							end
							
							local weaponname = InvManager.GetInventoryItemParameter(info.itemtype, "Name")
							if weaponname then
								tooltiptext = tooltiptext.." - "..weaponname
							end
							
							if info.controller then
								local name = GetPlayerName(info.controller) or "Controlled by unknown?"
								local armorstr = "Armor: "..info.armorpercent.."%"
								tooltiptext = tooltiptext.."\n"..name.."\n"..armorstr
							else
								local desc = info.isdead and "Destroyed" or "Active"
								local armorstr = info.isdead and "" or "\nArmor: "..info.armorpercent.."%"
								tooltiptext = tooltiptext.."\n"..desc..armorstr
							end
							ShowTooltip(self.x + self.w, -self.y, tooltiptext)
						end,
						leavewindow_cb = function(self)
							HideTooltip()
						end,
					}
					
					turretdata.iconcontainer = iup.zbox{cx=0, cy=0, turretdata.icon, turretdata.selimage, turretdata.button, expand="NO", all="YES"}
					
					iup.Append(iconcontainer, turretdata.iconcontainer)
					
					turretdata.icon.visible = "NO"
					turretdata.iconcontainer:map()
					turretdata.icon.visible = "YES"
					
					self.iconlist[turretinfo.itemid] = turretdata
					
					self:Update(turretinfo.itemid, turretinfo)
				end
			end
		end
		
		function box:UpdateIconPositions()
			if not self.iconlist then
				self:SetupIcons()
			end
			
			local invid = InvManager.InventoryItemToType(this.shipname or GetStationType())
			if not invid then return end
			local prototype = InvManager.GetInventoryItemPrototype(invid)
			if not prototype then return end
			
			local viewx, viewy, vieww, viewh = tonumber(modelview.x), tonumber(modelview.y), tonumber(modelview.w), tonumber(modelview.h)
			
			for turretid, turretdata in pairs(self.iconlist) do
				local portinfo = prototype:GetPortInfo(turretid)
				if portinfo then
					local pos = portinfo.position
					
					if self.cutoff_func then
						turretdata.iconcontainer.visible = self.cutoff_func(pos) and "YES" or "NO"
					end
					
					local proj = modelview:ProjectPoint(pos)
					if proj then
						turretdata.iconcontainer.cx = math.floor(proj:x() - viewx - turretdata.size/2)
						turretdata.iconcontainer.cy = math.floor(proj:y() - viewy - turretdata.size/2)
					end
				end
			end
			
			iconcontainer.refresh = 1
		end
		
		function box:Update(turretid, turretinfo)
			if self.iconlist and self.iconlist[turretid] then
				local turretdata = self.iconlist[turretid]
				turretdata.info = turretinfo
				if turretinfo.isdead then
					turretdata.icon.bgcolor = "255 255 255 150 *"
				else
					local healthcolor = math.floor((turretinfo.armorpercent/100)*225)
					turretdata.icon.bgcolor = "255 "..healthcolor.." "..healthcolor.." 255 *"
				end
			end
		end
		
		box.view = modelview
		box.iconcontainer = iconcontainer
		box.name = name
		
		return box
	end

	turretviewer.modelviews = {}
	for i=1, 2 do
		turretviewer.modelviews[i] = turretviewer:CreateModelViewer()
	end

	function turretviewer:SetModelViewAngle(viewindex, angle, cutoff)
		if viewindex < 1 or viewindex > #self.modelviews then return end
		self.modelviews[viewindex].cutoff_func = cutoff
		self.modelviews[viewindex].view:SlerpOrientation(angle)
		-- self.modelviews[viewindex]
	end

	function turretviewer:SetModelViewName(viewindex, name)
		self.modelviews[viewindex].name.title = name
	end

	function turretviewer:Show(index)
		index = index or 1
		for i = 1, 2 do
			local viewindex = (index * 2 - 2) + i
			self:SetModelViewAngle(i, self.views[viewindex].angle, self.views[viewindex].cutoff)
			self:SetModelViewName(i, self.views[viewindex].name)
		end
	end

	function turretviewer:InsertFrame()
		local self2 = self
		
		self.selectbutton = iup.stationbutton{title="Select", active="NO", action=function()
			if self.selectedturretdata then
				PurchaseMerchandiseItem(self.selectedturretdata.info.itemid)
			end
		end}
		
		self.toggleturretvisible = iup.stationtoggle{title="Show turrets", value="ON", font=Font.H5, action=function(_self, val)
			for i,v in ipairs(self.modelviews) do
				self.showturrets = val == 1
				v.iconcontainer.visible = self.showturrets and "YES" or "NO"
			end
		end}
		
		local viewradiobuttons = iup.hbox{
			iup.stationradio{title="Top / Bottom", action=function(_, v) if v == 1 then self:Show(1) end end},
			iup.stationradio{title="Right / Left", action=function(_, v) if v == 1 then self:Show(2) end end},
			iup.stationradio{title="Front / Back", action=function(_, v) if v == 1 then self:Show(3) end end},
			gap=12,
			alignment="ACENTER",
		}
		
		self.frame = iup.stationsubsubframe{
			iup.vbox{
				iup.hbox{
					self.modelviews[1],
					self.modelviews[2],
				},
				iup.hbox{
					self.selectbutton,
					iup.fill{size=10},
					self.toggleturretvisible,
					iup.fill{},
					iup.radio{
						viewradiobuttons,
						value=viewradiobuttons[1],
					},
					alignment="ACENTER",
				},
			},
		}
		
		function self.frame.OnShow(this)
			self.visible = true
			if self.reset or table.empty(self.turretlist) then
				self:ReloadView()
			end
		end
		
		function self.frame.OnHide(this)
			self.visible = false
		end
		
		local turrettab = _turrettab
		self.turretlistui = turrettab[1]
		
		self.turretlistui:detach()
		
		self.zbox = iup.zbox{self.turretlistui, self.frame, all="NO"}
		
		self.turretlistui.OnShow = turrettab.OnShow
		self.turretlistui.OnHide = turrettab.OnHide
		
		function turrettab.OnShow(this)
			if self.zbox.value.OnShow then
				if self.zbox.value == self.turretlistui then
					return self.zbox.value.OnShow(turrettab)
				else
					return self.zbox.value:OnShow()
				end
			end
		end
		
		function turrettab.OnHide(this)
			if self.zbox.value.OnHide then
				if self.zbox.value == self.turretlistui then
					return self.zbox.value.OnHide(turrettab)
				else
					return self.zbox.value:OnHide()
				end
			end
		end
		
		local tabradiobuttons = iup.hbox{
			iup.stationradio{title="Ship", action=function(_self, v)
				if v == 1 then
					self.zbox.value = self.frame			
					self.frame:OnShow()
					gkini.WriteInt("TurretViewer", "defaulttab", 1)
				else
					self.frame:OnHide()
				end
			end},
			iup.stationradio{title="List", action=function(_self, v)
				if v == 1 then
					self.zbox.value = self.turretlistui
					self.turretlistui.OnShow(turrettab)
					gkini.WriteInt("TurretViewer", "defaulttab", 2)
				else
					self.turretlistui.OnHide(turrettab)
				end
			end},
			alignment="ACENTER",
			gap=8,
		}
		
		self.box = iup.vbox{
			iup.stationsubsubframe{
				iup.hbox{
					iup.fill{},
					iup.radio{
						tabradiobuttons,
						value=tabradiobuttons[gkini.ReadInt("TurretViewer", "defaulttab", 1)],
					},
				},
			},
			self.zbox,
		}
		
		iup.Append(turrettab, self.box)
		
		self.zbox.value = gkini.ReadInt("TurretViewer", "defaulttab", 1) == 1 and self.frame or self.turretlistui
		turrettab:map()
		
		iup.Refresh(turrettab)
	end

	function turretviewer:ReloadView()
		self.reset = false
		
		local shipname = GetStationType()
		if shipname then
			self:SetModel(shipname)
		end
		
		self.turretnames = self.turretnames or GetTurretNamesByStationType(GetStationType())
		self.turretlist = {}
		for i=1, GetNumStationTurrets() do
			self.turretlist[i] = GetStationTurretInfo(i)
		end
		
		for i,v in ipairs(self.modelviews) do
			v:SetupIcons()
			v:UpdateIconPositions()
		end
	end

	function turretviewer:SetModel(shiptype)
		local factionid = GetStationFaction()
		
		self.shipname = shiptype
		self.meshinfo = {self:GetMeshInfo(shiptype)}
		
		if self.meshinfo[1] then
			for i,v in ipairs(self.modelviews) do
				v.view.value = self.meshinfo[2]..":"..self.meshinfo[1]
				v.view.fgcolor = ShipPalette_string[FactionColor[factionid] or 999] or "255 255 255"
				
				if v.view.currentangle then
					v.view:SetOrientation(v.view.currentangle)
				end
				
				v:SetupIcons()
				v:UpdateIconPositions()
			end
		end
	end
	
	function turretviewer:Update(index)
		for i,v in ipairs(self.modelviews) do
			v:Update(self.turretlist[index].itemid, self.turretlist[index])
		end
	end

	function turretviewer:OnEvent(event, ...)
		if event == "ENTERING_STATION" or event == "STATION_UPDATED" or event == "STATION_UPDATE_PRICE" then
			if event == "ENTERING_STATION" then
				self.turretnames = nil
				self.selectedturretdata = nil
				if self.selectbutton then
					self.selectbutton.active = "NO"
				end
			end
			
			if self.visible then
				self:ReloadView()
			else
				self.reset = true
			end
		elseif event == "STATION_TURRET_HEALTH_UPDATE" and self.turretlist then
			local nodeid, objectid, armorpercent, shieldamount = ...
			for index, v in pairs(self.turretlist) do
				if v.nodeid == nodeid and v.objectid == objectid then
					v.armorpercent = armorpercent
					self:Update(index)
					break
				end
			end
		elseif event == "LEAVING_STATION" then
			self.selectedturretdata = nil
			self.turretnames = nil
			if self.selectbutton then
				self.selectbutton.active = "NO"
			end
		end
	end

	RegisterEvent(turretviewer, "ENTERING_STATION")
	RegisterEvent(turretviewer, "STATION_UPDATED")
	RegisterEvent(turretviewer, "STATION_UPDATE_PRICE")
	RegisterEvent(turretviewer, "LEAVING_STATION")
	RegisterEvent(turretviewer, "STATION_TURRET_HEALTH_UPDATE")

	turretviewer:InsertFrame(_turrettab)
	turretviewer:SetModel("trident")
	turretviewer:Show(1)
	
	return turretviewer
end

TurretViewer[1] = TurretViewer.Create(CapShipTurretTab)
TurretViewer[2] = TurretViewer.Create(CapShipShiplessTurretTab)

